
*
* WinStoch.gms:        The basic winvest model, solved as a true, 3-stage stochastic
*               program, using the scenario-tree formulation where we have
*               decision variables for each node of the scenario tree.
*               Can use 4-scenario data from WinData.inc, or the scenario
*               generation code from WinGen.inc, by changing the include:
*
* SN, GAMS Summer Schools, 2003.
*

*$include "WinData.inc"
$include "WinGen.inc"
$include "WinNodes.inc"
$eolcom #

*
* This model has decision variables x(i,n,s) associated with each node
* in the scenario tree (and, of course, each instrument and scenario).
*
* Also, the data is associated with nodes, not scenarios, so have to create
* "mirror parameters", indexed by node, for yields etc. (names end in "N"):

  positive variables
    x(n,i) "holding of instrument i, out of node n",
    buy(n,i) "purchases of instrument i at node n",
    sell(n,i) "sales of instrument i at node n",
    final(n) "final value after node n",
    cash(n)  "cash after node n";

  variable ExpWealth "Final expected wealth, to be maximized"

* Constraints:
  equations CNST1(n), CNST2(n0, n, i), CASHC(n0,n), TOT_FINAL;
* -- Initial holdings sum to 100:

  CNST1(root(n0)) .. sum(i, x(n0, i)) + cash(n0) =E= 100;

* -- At each second-stage (non-root) node, we "receive" the initial (root)
*    investment times its yield, and can sell if desired. Can also buy.
*    This defines the holdings "out of" the node:

  CNST2(ancestor(n0,n),i) ..
     x(n0,i) * yieldN(i,n) +             # Coming into node n (from its ancestor, n0)
     buy(n, i)
                 =E=
     sell(n, i) +
     x(n, i) $ (not leaf(n));            # Going out of node n, except leaves.

* Can't buy at a leaf node, can't sell at the root node:
  buy.fx(leaf(n), i)  = 0;
  sell.fx(root(n), i) = 0;

  CASHC(ancestor(n0,n)) ..               # Keep track of cash position
     cash(n0) * cash_yieldN(n) +
     sum(i, sell(n, i)) * (1-transcost)
                 =E=
     sum(i, buy(n, i))  +
     liabN(n) $ liabs +
     cash(n);                            # cash out of leaf nodes are final positions

  TOT_FINAL ..                           # Total final position:
     ExpWealth =E= sum( n $ leaf(n), cash(n) ) / card(Cscen);


  model WinStoch /all/;
  solve WinStoch maximizing ExpWealth using lp;


