$TITLE ClearLake model: page 44, Birge & Louveaux

SET     p       Precipitation levels in each month /low, normal, high/,
        t       Time periods /dec,jan,feb,mar,apr/,
        baset(t) / dec /,
        w       Weather conditions /wet, dry/;
SET     tw(t,w)        'relates months to weather conditions' /
                (jan,feb).wet,
                (mar,apr).dry
                /;

SET     n       'nodes' / n1 * n121 /;
SET     root(n)        'root node' / n1 /;
SET     tn(t,n) 'map nodes to time periods';
ALIAS   (n,parent,child);
SET     anc(child,parent) 'ancestor mapping';
SET     np(n,p);
SET     leaf(n);

np(n,p)$[mod(ord(n)-2,card(p)) eq ord(p)-1] = yes;
np(root,p) = no;
* display np;

scalar tmp1, tmp2;
tmp1 = 0;
loop {t,
    tmp2 = (power[card(p), ord(t)]) / (card(p)-1);
    tn(t,n)$[ord(n) ge tmp1  and  ord(n) lt tmp2] = yes;
    tmp1 = tmp2;
};
* display tn;
anc(child,parent)$[floor((ord(child)+1)/card(p)) eq ord(parent)] = yes;
 display anc;
leaf(n)$[ord(n) gt (power[card(p), card(t)-1] - 1) / (card(p)-1)] =  yes;
 display leaf;

TABLE  delta(w,p)  'Changes in reservoir level for each season'
                        low     normal  high
                dry     -50     100     250
                wet     50      150     350     ;

PARAMETER       pr(p)  'Probability distribution' /
low     0.25,
normal         0.50,
high    0.25
/;

PARAMETER nprob(n)      'probability of being at any node';
nprob(root) = 1;
loop {anc(child,parent),
    nprob(child) = sum {np(child,p), pr(p)} * nprob(parent);
};
* display nprob;

* ndelta required for current scenRed implementation
PARAMETER ndelta(n)     'water delta at each node';
ndelta(n) = sum {(tw(t,w), np(n,p))$[tn(t,n)], delta(w,p)};
* display ndelta;

tmp1 = sum {leaf, nprob(leaf)};
abort$[abs(tmp1-1) gt 1e-8] "Error in tree: leaf probabilities do not sum to 1";


SCALAR floodCost        'K$/mm for amounts over flood level' / 10 /;
SCALAR lowCost          'K$/mm for amounts 250mm below flood level' / 5 /;
SCALAR l0               'initial water level' /100/;

VARIABLE                ec      'Expected value of cost';
POSITIVE VARIABLE       l(t,n)  'level of water in dam, EOP';
POSITIVE VARIABLE       r(t,n)  'mm released normally';
POSITIVE VARIABLE       f(t,n)  'mm of floodwater released';
POSITIVE VARIABLE       z(t,n)  'mm of water imported';

r.up(t,n) = 200;
* water level l is relative to 250mm below flood stage
l.up(t,n) = 250;
l.fx(baset,n) = l0;

SET     nn(n)           'nodes in reduced tree';
SET     sanc(child,parent) 'ancestor mapping for reduced tree';
SET     canc(child,parent) 'computed ancestor mapping for reduced tree';
PARAMETER snprob(n)     'probabilities for reduced tree';

EQUATIONS               ecdef, ldef(t,n);

ecdef..                ec =e= sum {tn(t,nn), snprob(nn) *
                             [floodCost * f(t,nn) + lowCost * z(t,nn)]};

ldef(tn(t,nn))$[not root(nn)]..
                l(t,nn) =e= sum {anc(nn,parent), l(t-1,parent)}
                         + ndelta(nn) + z(t,nn)
                         - r(t,nn) - f(t,nn);

model mincost / ecdef, ldef /;

$if set scenred $goto scenreduction
nn(n) = yes;
snprob(nn) = nprob(nn);
$goto donered

$label scenreduction
* now let's shrink the node set
parameter ScenRedParms(*);
parameter ScenRedReport(*);
ScenRedParms('num_leaves') =  sum {leaf, 1};
ScenRedParms('num_random') = 1;
ScenRedParms('num_nodes') = card(n);
ScenRedParms('num_time_steps') = card(t);
* at least one of the following two parameters is desired
ScenRedParms('red_num_leaves') =  sum {leaf, 1} ;
ScenRedParms('red_percentage') = 0.0;
* optional SCENRED input parameters
ScenRedParms('num_stages') = ScenRedParms('num_time_steps');
ScenRedParms('reduction_method') = 3;
ScenRedParms('where_random') = 1;
ScenRedParms('run_time_limit') = 60;
ScenRedParms('report_level') = 0;

execute_unload 'lakein.gdx',  ScenRedParms, n, anc, nprob, ndelta;
file opts / 'scenred.opt' /;
put opts;
put 'log_file =    lakelog.txt' /;
put 'input_gdx     lakein.gdx'  /;
put 'output_gdx =  lakeout.gdx' /;
putclose;
execute 'scenred scenred.opt';
execute_load   'lakeout.gdx', ScenRedReport, snprob=red_prob,
                              sanc=red_ancestor;
display scenredreport;
display snprob;

snprob(root) = 1;
nn(n) = snprob(n);
display nn;
* canc(leaf(child),parent)$(nn(child) and anc(child,parent)) = YES;
canc(anc(child,parent))$[nn(child)] = YES;
display canc, sanc;

SET     chk(child,parent);
chk(child,parent) = NO;
chk(canc(child,parent)) = YES;
chk(sanc(child,parent)) = NO;
display chk;
chk(child,parent) = NO;
chk(sanc(child,parent)) = YES;
chk(canc(child,parent)) = NO;
display chk;


tmp1 = sum {leaf(nn), snprob(leaf)};
display tmp1;
abort$[abs(tmp1-1) gt 1e-8] "Error in tree: leaf probabilities do not sum to 1";

$label donered

solve mincost using lp min ec;