$title Stochastic Programming
option limrow=0, limcol = 0, solprint=on;
$setglobal stages 4
$offlisting

set scenario /mb,ms,ok,ps,pb/;

parameter scenprob(scenario) /
mb 0.05
ms 0.25
ok 0.4
ps 0.25
pb 0.05 /;

parameter diff(scenario) /
mb -2
ms -1
ps 1
pb 2 /;

set i /1*9/;
alias (j,i);

parameter target(i) /
3 0.1
4 0.8
5 0.8
6 0.8
7 0.1
/;

parameter M(i,i,scenario);
loop {scenario,
    M(i+diff(scenario),i,scenario)$(target(i)) = 1;
};

* display M;


* The ordering of the following is used subsequently. Don't change it.
* time-2 -> n-30;  time-3 -> n-155;  time-4 -> n-780; time-5 -> n-3905
* time-6 -> n-19530
set t / time-0, time-1*time-%stages% /;
set n 'Nodes' / n-0,n-1*n-19530 /;
set nn(n)	'nodes actually in the model';
* set n 'Nodes' / n-0,n-1*n-3905 /;
alias (n,parent,child);
set sn(n) 'subset of nodes defining reduced tree'

set tn(t,n) time node mapping,
    tree(child,parent),
    event(n,scenario);

scalar startt, endt, numnodes;
endt = 1;
loop(t$(not sameas(t,'time-0')),
  startt = endt;
  endt = startt + power(card(scenario),ord(t)-1);
  tn(t,n)$((ord(n) le endt) and
           (startt lt ord(n))) = yes;
);

numnodes = endt;
abort$(card(n) lt numnodes) "Not enough nodes";
nn(n) = [ord(n) <= numnodes];

startt = 1;
loop(n$(ord(n) le numnodes-power(card(scenario),card(t)-1)),
  endt = startt + card(scenario);
  tree(child,n)$(ord(child) gt startt and ord(child) le endt) = yes;
  startt = endt;
);

loop(scenario,
  event(nn(n),scenario)$(ord(scenario) eq mod(ord(n)-2,card(scenario))+1)
    = yes;
);
event('n-0',scenario) = no;

parameter stpdat(n);
parameter stddat(n);
stpdat(nn) = sum(event(nn,scenario), scenprob(scenario));
stddat(nn) = sum(event(nn,scenario), diff(scenario));
parameter allM(i,i,n);
allM(i,j,nn) = sum{event(nn,scenario), M(i,j,scenario)};

parameter prob(n)  node probability;
parameter sprob(n)  'node probability, reduced subtree';

prob('n-0') = 1;
loop(t$(ord(t) gt 1),
  prob(n)$tn(t,n) = sum(tree(n,parent), stpdat(n)*prob(parent));
);

variables obj, delta(i,n), x(i,t,n), u(i,t,n);
positive variables u;

equations state(i,t,n), cost, upcost(i,n), downcost(i,n);

cost..
obj =e= sum(i,sum(tn('time-%stages%',sn),sprob(sn)*delta(i,sn)));

state(i,tn(t,sn))..
  x(i,t,sn) =e= sum(tree(sn,parent),  x(i,t-1,parent)
               + sum {j, allM(i,j,sn) * u(j,t-1,parent)});

upcost(i,sn)$tn('time-%stages%',sn)..
delta(i,sn) =g= x(i,'time-%stages%',sn) - target(i);

downcost(i,sn)$tn('time-%stages%',sn)..
delta(i,sn) =g= target(i) - x(i,'time-%stages%',sn);

$if set scenred $goto scenreduction
sn(n) = yes;
sprob(sn) = prob(sn);
$goto donered
 
$label scenreduction
* now let's shrink the node set
parameter ScenRedParms(*);
parameter ScenRedReport(*);
ScenRedParms('num_leaves') = sum {tn('time-%stages%',n), 1};
ScenRedParms('num_random') = card(scenario) * sum {i$(target(i)), 1};
* ScenRedParms('num_random') = sum {(i,j)$[sum{n$allm(i,j,n), 1}], 1};
ScenRedParms('num_nodes')  = numnodes;
ScenRedParms('num_time_steps') = card(t);
* at least one of the following two parameters is required
ScenRedParms('red_num_leaves') = ScenRedParms('num_leaves') * 0.15;
* ScenRedParms('red_percentage') = 0.5;
ScenRedParms('num_stages') = ScenRedParms('num_time_steps');
ScenRedParms('reduction_method') = 1;
ScenRedParms('where_random') = 1;
ScenRedParms('run_time_limit') = 30;
ScenRedParms('report_level') = 0;
display ScenRedParms;

$onlisting
sn(n)$[ord(n) le numnodes] = yes;
execute_unload 'kandin.gdx',  ScenRedParms, sn, tree, prob, allM
file opts / 'scenred.opt' /;
put opts;
put 'log_file =    kandlog.txt' /;
put 'input_gdx     kandin.gdx'  /;
put 'output_gdx =  kandout.gdx' /;
putclose;

scalar rc;
$if     %system.filesys% == UNIX execute 'rm -f kandout.gdx';
$if not %system.filesys% == UNIX execute 'del   kandout.gdx';
execute 'scenred scenred.opt';
rc = errorlevel;
abort$rc "Return code from scenred was nonzero : ", rc;
execute_load   'kandout.gdx', ScenRedReport, sprob=red_prob;
display scenredreport;
sprob('n-0') = 1;
sn(n) = sprob(n);


$label donered

scalar theta /1.0/;
x.fx(i,'time-0','n-0') = 0;
u.up(i,'time-0','n-0') = theta;
u.up(i,tn(t,sn)) = theta;

model stoch /all/;
stoch.iterlim = 10000000;
stoch.optfile = 1;

solve stoch using lp minimizing obj;
