$include winstoch.gms

$eolcom #

* Now we prepare to run ScenRed

* This includes some sets & parameters used for scenred I/O
$libinclude scenred.gms

* First we analyse the GAMS program of the Winvest model.
* Since the initial scenarios and a number of input parameters have to be passed
* to SCENRED one must identify the corresponding components of the GAMS model
* and create or calculate them if they do not already exist.

* The model contains card(Cscen)=8 scenarios. They form a scenario tree with
* card(Cnodes)=12 nodes.
ScenRedParms('num_leaves') =  card(Cscen);
ScenRedParms('num_nodes') = card(n);
ScenRedParms('num_time_steps') = card(Ctime);
* To each node n there are assigned card(i)+2 random
* values: yieldN(i,n), cash_yieldN(n) and liabN(n).
ScenRedParms('num_random') = card(i)+2;
* The structure of the scenario tree is given by the dynamic set ancestor.

* The only missing component is a parameter for the node
* probabilities. Since the variable ExpWealth represents the final expected
* wealth of our portfolio, the scenario probabilites can be found in the
* objective TOT_FINAL: the scenarios have equal probabilities 1/card(Cscen).
*
* The node probabilities of the initial scenario tree will be contain in the
* parameter prob. We will define the probabilities of the leaves; the
* probabilities of the remaining nodes are computed during the SCENRED run
parameter prob(n)  "node probability of the initial tree";
prob(leaf(n))=1/card(Cscen);

* Since  WINSTOCH cannot handle varying scenario or node probabilities,
* we add the missing components to reformulate the GAMS model above.
set sn(n) "subset of nodes in reduced subtree";
parameter  sprob(n) "node probability in reduced tree";

*  The following parameters control the tree output from ScenRed;
*  at least one of the following two parameters is required
  ScenRedParms('red_num_leaves') = card(CScen)/2;
*  ScenRedParms('red_percentage') = 0.5;

* set up the scenred options file
file opts / 'scenred.opt' /;
putclose opts  'log_file    SRLogWinStoch.txt'
             / 'input_gdx   SRInWinStoch.gdx'
             / 'output_gdx  SROutWinStoch.gdx';

scalar    rc;

* Reduce the scenario tree
execute_unload 'SRInWinStoch.gdx',  ScenRedParms, n, ancestor, prob, yieldN,cash_yieldN,liabN ;
execute 'scenred scenred.opt %system.redirlog%';
        rc = errorlevel;
        abort$rc "Return code from scenred was nonzero : ", rc;
execute_load   'SROutWinStoch.gdx', ScenRedReport, sprob=red_prob;

* The node set of the reduced scenario tree
sn(n)=sprob(n);



* The decision variables associated with each node in the reduced
* scenario tree are x(i,sn(n)),  buy(sn(n),i), sell(sn(n),i), final(sn(n))
* and cash(sn(n)). Hence, the decision variables are already declared.
*

* We now reformulate the constraints CONST1, CONST2, CASHC and TOT_FINAL to
* depend on sn.
equations SRCNST1(n), SRCNST2(n0, n, i), SRCASHC(n0,n), SRTOT_FINAL;
* -- Initial holdings sum to 100:

  SRCNST1(root(sn(n0))) .. sum(i, x(n0, i)) + cash(n0) =E= 100;

* -- At each second-stage (non-root) node, we "receive" the initial (root)
*    investment times its yield, and can sell if desired. Can also buy.
*    This defines the holdings "out of" the node:

  SRCNST2(ancestor(n0,sn),i) ..
     x(n0,i) * yieldN(i,sn) +         # Coming into node sn(n) (from its ancestor, n0)
     buy(sn, i)
                 =E=
     sell(sn, i) +
     x(sn, i) $ (not leaf(sn));            # Going out of node sn, except leaves.

  SRCASHC(ancestor(n0,sn)) ..               # Keep track of cash position
     cash(n0) * cash_yieldN(sn) +
     sum(i, sell(sn, i)) * (1-transcost)
                 =E=
     sum(i, buy(sn, i))  +
     liabN(sn) $ liabs +
     cash(sn);                            # cash out of leaf nodes are final positions

  SRTOT_FINAL ..                           # Total final position:
     ExpWealth =E= sum( sn $ leaf(sn), sprob(sn)*cash(sn) );

* Can't buy at a leaf node, can't sell at the root node:
  buy.fx(leaf(sn), i)  = 0;
  sell.fx(root(sn), i) = 0;

model SRWinStoch /SRCNST1,SRCNST2,SRCASHC,SRTOT_FINAL/;
solve SRWinStoch maximizing ExpWealth using lp;
